/**
 * Copyright DataStax, Inc.
 *
 * Please see the included license file for details.
 */
package com.datastax

import java.util.UUID
import java.net.URI

object HttpReceiverCases {

  sealed trait ReceiverClass extends Serializable

  val DEFAULT_RECEIVERS = 1
  val DEFAULT_PORT = 9999

  case class Config(
    numDstreams: Int = DEFAULT_RECEIVERS,
    port: Int = DEFAULT_PORT) extends ReceiverClass

  /**
   * A representation of the C* Timeline table
   * @param timesegment A 10 second window of time which will be used as a bin for partitioning requests
   * @param url The URI requested on the server
   * @param t_uuid A timeuuid encapsulating the time at which the request was received by the server
   * @param http_method The HTTP method used for this request (POST,GET,... )
   * @param headers All of the header data associated with this request
   * @param body The body of this request
   */
  case class timelineRow(
    timesegment: Long,
    url: String,
    t_uuid: UUID,
    http_method: String,
    headers: Map[String, String],
    body: String) extends ReceiverClass

  /**
   * A representation of the C* method_agg table
   * @param time Time of the batch which is being analyzed
   * @param url The uri requested
   * @param http_method The Http method used
   * @param count The number of times this particular uri and method was requested during the "time" batch
   */
  case class methodAggRow(
    time: Long,
    url: String,
    http_method: String,
    count: Long) extends ReceiverClass

  /**
   * A representation of the C* sorted_urls table
   * @param time Time of the batch which is being analyzed
   * @param url The uri requested
   * @param count Number of requests during this batch
   */
  case class sortedUrlRow(
    time: Long,
    url: String,
    count: Long) extends ReceiverClass

  /**
    * Serializable version of the HttpResponses generated by
    * the HttpServer
    * @param timeuuid
    * @param http_method
    * @param headers
    * @param uri
    * @param body
    */
  case class HttpRequest(
    timeuuid: UUID,
    http_method: String,
    headers: Map[String, List[String]],
    uri: URI,
    body: String) extends ReceiverClass

}
