/*
 * Copyright DataStax, Inc.
 *
 * Please see the included license file for details.
 */
package com.datastax.dse.demos.solr.commands;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;

import javax.annotation.Nullable;

import com.google.common.base.Splitter;

import com.datastax.dse.demos.solr.CommandLine;
import com.datastax.dse.demos.solr.clients.ConcurrentBatchUpdateSolrServer;
import com.datastax.dse.demos.solr.SolrStress;
import com.datastax.dse.demos.solr.Utils;
import com.datastax.dse.demos.solr.clients.Clients;
import com.datastax.dse.demos.solr.stats.Metrics;

import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListenableFuture;
import org.apache.solr.client.solrj.SolrQuery;
import org.apache.solr.client.solrj.util.ClientUtils;
import org.apache.solr.common.SolrDocument;
import org.apache.solr.common.SolrDocumentList;
import org.apache.solr.common.SolrInputDocument;

public class HttpUpdateCmdRunner extends CmdRunner
{
    public final static String CMD_STRING_ID = "HTTPUPDATE";
    public final static String UPDATE_ALL_STRING_FIELDS_ID = "ALLSTR";
    private final ConcurrentBatchUpdateSolrServer client;

    HttpUpdateCmdRunner(CommandLine.Params params, Metrics metrics)
    {
        super(metrics);
        this.client = Clients.getConcurrentBatchUpdateSolrServer(params.urls, params.clients, params.concurrency, params.consistencyLevel);
    }

    @Override
    public void runCommand(List<String> cmd2Run) throws Throwable
    {
        // cmd2Run[0] = update command [1]= csv list of fields to update or 'ALLSTR' [2]=query producing resultset to be updated
        SolrQuery query = new SolrQuery();

        // Take escaping into account
        for (String nameValuePair : cmd2Run.get(2).split("(?<!&)&(?!&)"))
        {
            nameValuePair = nameValuePair.replace("&&", "&");
            String[] nameValuePairArray = nameValuePair.split("(?<!=)=(?!=)");
            query.set(nameValuePairArray[0], nameValuePairArray[1].replace("==", "="));
        }

        // Get only num found and only one doc when querying
        query.setRows(1);
        start = System.nanoTime();
        long numFound = client.query(query).getResults().getNumFound();
        metrics.update(Metrics.StatsType.DOC_COUNT, numFound, true);

        if (numFound == 0)
        {
            System.err.println(this.getClass().getName() + " could not find anything to update for query: " + query);
            return;
        }

        query.setStart(ThreadLocalRandom.current().nextInt((int) numFound));
        SolrDocumentList results = client.query(query).getResults();

        List<String> updatableFields = Splitter.on(',').trimResults().omitEmptyStrings().splitToList(cmd2Run.get(1));
        int updates = 0;
        List<SolrInputDocument> toUpdate = new ArrayList<SolrInputDocument>();
        for (SolrDocument doc : results)
        {
            for (String fieldName : doc.getFieldNames())
            {
                if (updatableFields.contains(fieldName) || updatableFields.contains(UPDATE_ALL_STRING_FIELDS_ID))
                {
                    Object docValue = doc.getFieldValue(fieldName);
                    if (docValue instanceof String)
                    {
                        doc.setField(fieldName, "updated" + ++updates + " " + docValue);
                    }
                }
            }

            // TODO: This is an in-lined deprecated method from Solr...super hacky.
            SolrInputDocument inputDoc = new SolrInputDocument();
            
            for (String name : doc.getFieldNames()) 
            {
                inputDoc.addField( name, doc.getFieldValue(name), 1.0f );
            }
            
            toUpdate.add(inputDoc);
        }

        if (toUpdate.isEmpty())
        {
            System.err.println(this.getClass().getName() + " No string fields to update found for query: " + query + " still docs have been read and written to with their original values.");
        }
        else
        {
            final long startNanos = start;
            ListenableFuture<Object> resultFuture = client.add(toUpdate);
            FutureCallback<Object> callback = new FutureCallback<Object>()
            {

                @Override
                public void onSuccess(@Nullable Object result)
                {
                    metrics.update(Metrics.StatsType.OP_LATENCY, startNanos, System.nanoTime(), true);
                }

                @Override
                public void onFailure(Throwable t)
                {
                    SolrStress.logException(t);
                    metrics.update(Metrics.StatsType.OP_LATENCY, startNanos, System.nanoTime(), false);
                }
            };
            Futures.addCallback(resultFuture, callback, Utils.DIRECT_EXECUTOR);
        }
    }
}
