/*
 * Copyright DataStax, Inc.
 *
 * Please see the included license file for details.
 */
package com.datastax.dse.demos.solr.commands;

import java.net.MalformedURLException;
import java.util.List;

import javax.annotation.Nullable;

import com.datastax.dse.demos.solr.CommandLine;
import com.datastax.dse.demos.solr.clients.ConcurrentBatchUpdateSolrServer;
import com.datastax.dse.demos.solr.SolrStress;
import com.datastax.dse.demos.solr.Utils;
import com.datastax.dse.demos.solr.clients.Clients;
import com.datastax.dse.demos.solr.probes.IndexingLatencyProbe;
import com.datastax.dse.demos.solr.stats.Metrics;

import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListenableFuture;
import org.apache.solr.common.SolrInputDocument;

public class HttpWriteCmdRunner extends CmdRunner
{
    public static final String CMD_STRING_ID = "HTTPWRITE";
    private final CommandLine.Params params;
    private final ConcurrentBatchUpdateSolrServer client;
    private final IndexingLatencyProbe indexingLatencyProbe;

    HttpWriteCmdRunner(CommandLine.Params params, Metrics metrics, IndexingLatencyProbe indexingLatencyProbe) throws MalformedURLException
    {
        super(metrics);
        this.params = params;
        this.client = Clients.getConcurrentBatchUpdateSolrServer(params.urls, params.clients, params.concurrency, params.consistencyLevel);
        this.indexingLatencyProbe = indexingLatencyProbe;
    }

    @Override
    public void runCommand(List<String> cmd2Run) throws Throwable
    {
        SolrInputDocument doc = new SolrInputDocument();
        for (String fieldData : cmd2Run.subList(1, cmd2Run.size()))
        {
            // Take character doubling escaping into account
            String[] explodedField = fieldData.replace("&&", "&").split("(?<!=)=(?!=)");
            if (explodedField.length > 1)
            {
                String fieldName = explodedField[0];
                String fieldValue = Utils.runCmdNotationSubstitutions(explodedField[1].replace("==", "="), params.random.get());
                doc.addField(fieldName, fieldValue);
            }
        }
        start = System.nanoTime();
        final long startNanos = start;
        ListenableFuture<Object> resultFuture = client.add(doc);
        FutureCallback<Object> callback = new FutureCallback<Object>()
        {
            @Override
            public void onSuccess(@Nullable Object result)
            {
                long nowNanos = System.nanoTime();
                metrics.update(Metrics.StatsType.OP_LATENCY, startNanos, nowNanos, true);
                indexingLatencyProbe.maybeProbe((String)doc.getFieldValue(params.indexingLatencyProbeColumn), nowNanos);
            }

            @Override
            public void onFailure(Throwable t)
            {
                SolrStress.logException(t);
                metrics.update(Metrics.StatsType.OP_LATENCY, startNanos, System.nanoTime(), false);
            }
        };
        Futures.addCallback(resultFuture, callback, Utils.DIRECT_EXECUTOR);
    }

}
