/*
 * Copyright DataStax, Inc.
 *
 * Please see the included license file for details.
 */

package com.datastax.dse.demos.solr.load;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Deque;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Future;

import javax.annotation.Nullable;

import com.google.common.util.concurrent.FutureCallback;
import com.google.common.util.concurrent.Futures;

import com.datastax.dse.demos.solr.probes.IndexingLatencyProbe;
import org.apache.cassandra.utils.Throwables;
import org.apache.solr.common.SolrInputDocument;

import com.google.common.util.concurrent.ListenableFuture;
import com.datastax.dse.demos.solr.CommandLine;
import com.datastax.dse.demos.solr.clients.ConcurrentBatchUpdateSolrServer;
import com.datastax.dse.demos.solr.SolrStress;
import com.datastax.dse.demos.solr.clients.Clients;
import com.datastax.dse.demos.solr.stats.Metrics;

public class HTTPLoadStrategy extends LoadStrategy
{
    private final ConcurrentBatchUpdateSolrServer client;
    public HTTPLoadStrategy(CommandLine.Params params, Metrics metrics, InputLoader reader, IndexingLatencyProbe indexingLatencyProbe) {
        super(params, metrics, reader, indexingLatencyProbe);
        client = Clients.getConcurrentBatchUpdateSolrServer(params.urls, params.clients, params.concurrency, params.consistencyLevel);
    }

    @Override
    public void execute()
    {
        Throwable mergedException = null;
        Deque<Future<Object>> futures = new ArrayDeque<>();
        try
        {
            List<Map<String, String>> fieldss = new ArrayList<>(Clients.CONCURRENT_UPDATE_BATCH_SIZE);
            for (Map<String, String> fields : reader.mapAdapter())
            {
                if (fieldss.size() >= Clients.CONCURRENT_UPDATE_BATCH_SIZE)
                {
                    futures.offer(addDocs(fieldss));
                    fieldss = new ArrayList<>();
                }
                fieldss.add(fields);

                Future<Object> fut = null;
                while ((fut = futures.peek()) != null && fut.isDone())
                {
                    futures.poll();
                    metrics.logExecutedCommands(1);
                }
            }

            if (!fieldss.isEmpty())
                addDocs(fieldss);
        }
        catch (Throwable thrown)
        {
            mergedException  = Throwables.merge(mergedException, thrown);
        }
        finally
        {
            //Wait for everything to complete
            for (Future<Object> future : futures)
            {
                try
                {
                    future.get();
                }
                catch (Throwable t)
                {
                    Throwables.merge(mergedException, t);
                }
            }

            Throwables.maybeFail(mergedException);
        }
    }

    private Future<Object> addDocs(Collection<Map<String, String>> fieldss) throws Exception
    {
        final int fieldssSize = fieldss.size();

        List<SolrInputDocument> docs = new ArrayList<>(fieldssSize);
        List<String> probeIds = new ArrayList<>(fieldssSize);

        final long start = System.nanoTime();
        for (Map<String, String> fields : fieldss)
        {
            SolrInputDocument doc = new SolrInputDocument();
            for (Map.Entry<String, String> entry : fields.entrySet())
            {
                doc.addField(entry.getKey(), entry.getValue());
            }
            docs.add(doc);
            probeIds.add(fields.get(reader.pkeyColumnName()));
        }
        final ListenableFuture<Object> resultFuture = client.add(docs);
        Futures.addCallback(resultFuture, new FutureCallback<Object>() {

            @Override
            public void onSuccess(@Nullable Object result)
            {
                long nowNanos = System.nanoTime();
                for (int ii = 0; ii < fieldssSize; ii++)
                {
                    metrics.update(Metrics.StatsType.OP_LATENCY, start, nowNanos, true);
                }
                for (String probeId : probeIds)
                {
                    indexingLatencyProbe.maybeProbe(probeId, nowNanos);
                }
            }

            @Override
            public void onFailure(Throwable t)
            {
                long nowNanos = System.nanoTime();
                for (int ii = 0; ii < fieldssSize; ii++)
                {
                    metrics.update(Metrics.StatsType.OP_LATENCY, start, nowNanos, false);
                }
                SolrStress.logException(t);
            }
        });

        return resultFuture;
    }
}
